let silica_components = {};

class Silica {
    components = []

    static initComponents() {
        let componentEls = document.querySelectorAll("[silica\\:initial-data], [silica\\:lazy]");

        componentEls.forEach((componentEl) => {
            let component_id = componentEl.getAttribute("silica:id");

            if (!silica_components.hasOwnProperty(component_id)) {
                silica_components[component_id] = new SilicaComponent(componentEl);
            }
        });
    }

    static getNearestComponent(elem) {
        for (; elem && elem !== document; elem = elem.parentNode) {
            if (elem.getAttribute("silica:id")) {
                return {
                    el: elem,
                    id: elem.getAttribute("silica:id"),
                    name: elem.getAttribute("silica:name"),
                    component: silica_components[elem.getAttribute("silica:id")]
                };
            }
        }
        return false;
    }

    static broadcastEvent() {
    }

    static data_get(data, path) {
        return path.split('.').reduce((o, i) => o?.[i], data);
    }

    static data_key_exists(data, key) {
        return key.split('.').every(i => (data = data[i]) !== undefined);
    }
}


document.addEventListener("DOMContentLoaded", function () {
    Silica.initComponents();
});


// Wire in silica's magic methods to Alpine
document.addEventListener('alpine:init', () => {
    window.Alpine.magic('set', (el, {Alpine}) => (key, value) => {
        try {
            const {component} = Silica.getNearestComponent(el)
            if (component) {
                component.setProperty(key, value)
            }
        } catch (e) {
            console.error("Error calling $set Alpine magic method", e)
        }
    })
    window.Alpine.magic('call', (el, {Alpine}) => (method, args) => {
        try {
            const {component} = Silica.getNearestComponent(el)
            if (component) {
                component.callMethod(method, args)
            }
        } catch (e) {
            console.error("Error calling $call Alpine magic method", e)
        }
    })
})

class SilicaComponent {
    el;
    id;
    name;
    state = {};

    last_request_timestamp = null;

    boundMethods = {
        handleModelInputEvent: this.handleModelInputEvent.bind(this),
        handleClickEvent: this.handleClickEvent.bind(this)
    };

    ACTION_TYPE_CALL_METHOD = "call_method";
    ACTION_TYPE_SET_PROPERTY = "set_property";
    ACTION_TYPE_EVENT = "event";

    constructor(componentEl) {
        const encodedData = componentEl.getAttribute("silica:initial-data");
        const decodedData = decodeURIComponent(encodedData);
        const snapshot = JSON.parse(decodedData);

        this.el = componentEl
        this.id = componentEl.getAttribute("silica:id");
        this.name = componentEl.getAttribute("silica:name");
        this.data = snapshot.data;

        if (componentEl.hasAttribute("silica:lazy")) {
            this.activateLazy()
            componentEl.removeAttribute("silica:lazy")
            return
        }

        if (this.processRedirections(snapshot.js_calls)) {
            return;
        }

        componentEl.removeAttribute("silica:initial-data");

        this.updateModelValues(snapshot.data);
        this.setQueryParams(snapshot);
        this.processJsCalls(snapshot.js_calls);
        this.processEvents(snapshot?.event_calls);
        this.setListeners()
    }

    updateModelValues(data) {
        let mthis = this
        this.querySelectorAll("[silica\\:model]").forEach((el) => {
            let modelName = el.getAttribute("silica:model");
            let modelValue = Silica.data_get(data, modelName);

            if (Silica.data_key_exists(data, modelName)) {
                if (el.type === 'checkbox') {
                    if (Array.isArray(modelValue)) {
                        el.checked = modelValue.includes(el.value);
                    } else {
                        el.checked = modelValue;
                    }
                } else {
                    el.value = modelValue;
                }
            }
        });
    }

    setQueryParams(snapshot) {
        if (snapshot["query_params"] && snapshot["query_params"].length > 0) {
            snapshot["query_params"].forEach((param) => {
                if (snapshot.data?.[param.key] !== undefined) {

                    const url_key = param?.as || param.key

                    if (param.visible) {
                        this.setQueryParam(url_key, snapshot.data[param.key]);
                    } else {
                        this.removeQueryParam(url_key);
                    }
                }
            });
        }
    }

    setQueryParam(paramKey, paramValue) {
        const url = new URL(window.location);

        if (Array.isArray(paramValue)) {
            url.searchParams.delete(paramKey);
            paramValue.forEach(value => {
                url.searchParams.append(paramKey, value);
            });
        } else {
            url.searchParams.set(paramKey, paramValue);
        }

        window.history.pushState({}, "", url);
    }

    removeQueryParam(paramKey) {
        const url = new URL(window.location);
        url.searchParams.delete(paramKey)
        window.history.pushState({}, "", url);
    }

    processJsCalls(calls = []) {
        calls.forEach((call) => {
            const fn = call.fn;
            const args = call?.args;

            if (fn === '_silicaDispatchBrowserEvent') {
                this._silicaDispatchBrowserEvent(...args);
                return;
            }
            window[fn](...args);
        });
    }

    _silicaDispatchBrowserEvent(name, payload) {
        window.dispatchEvent(
            new CustomEvent(name, {
                detail: payload,
                bubbles: true,
                // Allows events to pass the shadow DOM barrier.
                composed: true,
                cancelable: true,
            })
        )
    }

    processEvents(events = []) {
        events.forEach((event) => {
            if (event.type === "emit") {
                // emit to all silica_components on page
                Object.values(silica_components).forEach((component) => {
                    const action = {
                        type: this.ACTION_TYPE_EVENT,
                        event_name: event.name,
                        payload: event.payload
                    };

                    component.sendMessage(action);
                });
            } else if (event.type === "emit_to") {
                // emit to all silica_components on page
                Object.values(silica_components)
                    .filter((component) => component.name === event.component_name)
                    .forEach((component) => {
                        const action = {
                            type: this.ACTION_TYPE_EVENT,
                            event_name: event.name,
                            payload: event.payload
                        };

                        component.sendMessage(action);
                    });
            }
        });
    }

    callMethod(name, args) {
        const action = {
            type: this.ACTION_TYPE_CALL_METHOD,
            method_name: name,
            args: args
        };

        this.sendMessage(action);
    }

    activateLazy() {
        const action = {
            type: "activate_lazy"
        };

        this.sendMessage(action);
    }

    setProperty(name, value) {
        if (name.includes('.')) {
            // Check the property exists
            let parts = name.split('.')
            let obj = this.data
            let i
            for (i = 0; i < parts.length - 1; i++) {
                if (obj.hasOwnProperty(parts[i])) {
                    obj = obj[parts[i]]
                } else {
                    console.error(`Property ${parts[i]} does not exist on component data`)
                    return
                }
            }
        }

        const action = {
            type: this.ACTION_TYPE_SET_PROPERTY,
            name: name,
            value: value
        };

        this.sendMessage(action);
    }

    sendMessage(action) {
        if (!this.id || !this.name) {
            console.error(
                "No Silica component element found when processing silica:click"
            );
            return;
        }

        // Initiate request concurrency checking
        this.last_request_timestamp = new Date().getTime();
        let current_request_timestamp = this.last_request_timestamp;

        // Show any silica:loading elements
        this.showLoaders(action);

        const params = {
            name: this.name,
            id: this.id,
            actions: [action]
        };

        // Send the POST request using fetch
        fetch("/silica/message", {
            method: "POST",
            headers: {
                "Content-Type": "application/json"
            },
            body: JSON.stringify(params)
        })
            .then((response) => response.json()) // assuming server responds with json
            .then((response) => {
                if (this.last_request_timestamp > current_request_timestamp) {
                    console.log('Request was superseded by a newer request')
                    return;
                }

                let snapshot = response.snapshot;

                if (this.processRedirections(snapshot.js_calls)) {
                    return;
                }

                this.updateDomWithAlpineMorph(response.html);

                Silica.initComponents(); // Pick up and init any new silica_components nested in others

                this.data = snapshot.data
                this.updateModelValues(snapshot.data);
                this.setQueryParams(snapshot);
                this.processJsCalls(snapshot.js_calls);
                this.processEvents(snapshot.event_calls);

                this.hideLoaders();
            })
            .catch((error) => {
                this.hideLoaders();
                console.error("Error:", error);
            });
    }

    // Performance TODO list
    // - In an init function, cache silica:models in the component

    /*
    find lazy silica_components
    call the activation
     */
    initLazyComponents() {
        const lazyComponentEls = this.querySelectorAll("[silica\\:lazy]");

        lazyComponentEls.forEach((componentEl) => {
            let component = silica_components.find(c => c.id === componentEl.getAttribute('silica:id'))

            if (!component) {
                console.error('Component for lazy activation not found in silica_components array')
                return
            }

            component.activateLazy()
        });
    }

    handleModelInputEvent(event) {
        const el = event.currentTarget;
        const modelName = el.getAttribute("silica:model");
        let currentValue = Silica.data_get(this.data, modelName);

        if (el.type === 'checkbox') {
            // Check if we have a multi-value checkbox
            if (Array.isArray(currentValue)) {

                if (!el.hasAttribute('value') || el.getAttribute('value') === '') {
                    console.error(`silica:model for checkbox ${modelName} is set to an array but the checkbox does not have a valid value attribute`)
                    return
                }

                let value = el.checked ? el.getAttribute('value') : false;

                if (value) {
                    currentValue.push(value)
                } else {
                    let valueToRemove = el.getAttribute('value') || true
                    currentValue = currentValue.filter(v => v !== valueToRemove)
                }

                this.setProperty(modelName, currentValue);
            } else {
                let value = el.checked ? (el.getAttribute('value') || true) : false;
                this.setProperty(modelName, value);
            }
        } else {
            this.setProperty(modelName, el.value);
        }
    }

    handleClickEvent(event) {
        const el = event.currentTarget;

        let instance = this

        /**
         * Handle a silica:click event
         * methodOrExpressionString can be a method name or a string expression, i.e. something = 1
         * @param event
         * @param methodOrExpressionString
         * @param prevent
         */
        function silicaClick(event, methodOrExpressionString, prevent = false) {
            if (prevent) {
                event.preventDefault();
            }

            // Check if string function call has args
            if (methodOrExpressionString.includes("(")) {
                const methodNameParts = methodOrExpressionString.split("(");
                const methodName = methodNameParts[0];

                // also remove any quotes from left or right of each arg
                const args = methodNameParts[1]
                    .replace(")", "")
                    .split(",")
                    .map((arg) => arg.trim().replace(/^['"]+|['"]+$/g, ''));
                instance.callMethod(methodName, args);
            }
            // Check if it's a property update (e.g., "something=2")
            else if (methodOrExpressionString.includes("=")) {
                const [propertyName, value] = methodOrExpressionString.split("=");

                instance.setProperty(propertyName.trim(), value.trim().replace(/^['"]+|['"]+$/g, ''));
                // Assume the string is a function without parentheses
            } else {
                instance.callMethod(methodOrExpressionString);
            }
        }

        if (el.hasAttribute("silica:click")) {
            const methodOrExpressionString = el.getAttribute("silica:click");
            silicaClick(event, methodOrExpressionString);
        } else if (el.hasAttribute("silica:click.prevent")) {
            const methodOrExpressionString = el.getAttribute("silica:click.prevent");
            silicaClick(event, methodOrExpressionString);
        }
    }

    setListeners() {
        const models = this.querySelectorAll("[silica\\:model]");
        models.forEach((element) => {
            if (element.type === "checkbox") {
                element.addEventListener("change", this.boundMethods.handleModelInputEvent);
            } else {
                element.addEventListener("input", this.boundMethods.handleModelInputEvent);
            }
        });

        const clickables = this.querySelectorAll("[silica\\:click\\.prevent]");
        clickables.forEach((element) => {
            element.addEventListener("click", this.boundMethods.handleClickEvent);
        });
    }

    removeListeners() {
        const models = this.querySelectorAll("[silica\\:model]");
        models.forEach((element) => {
            element.removeEventListener("input", this.boundMethods.handleModelInputEvent);
        });

        const clickables = this.querySelectorAll("[silica\\:click\\.prevent]");
        clickables.forEach((element) => {
            element.removeEventListener("click", this.boundMethods.handleClickEvent);
        });
    }

    updateDomWithAlpineMorph(html) {

        function isntElement(el) {
            return typeof el.hasAttribute !== 'function'
        }

        Alpine.morph(this.el, html, {
            updating: (el, toEl, childrenOnly, skip) => {
                if (isntElement(el)) return

                if (el.hasAttribute('silica:glued')) return skip()

                // Children will update themselves.
                if (el.hasAttribute('silica:id') && el.getAttribute('silica:id') !== this.id) return skip()
            },

            updated: (el) => {
                if (isntElement(el)) return
            },

            removing: (el, skip) => {
                if (isntElement(el)) return
            },

            removed: (el) => {
                if (isntElement(el)) return
            },

            adding: (el) => {
            },

            added: (el) => {
                if (isntElement(el)) return

                if (el.nodeName === 'SCRIPT') {
                    var script = document.createElement('script');
                    //copy over the attributes
                    [...el.attributes].forEach(attr => {
                        script.setAttribute(attr.nodeName, attr.nodeValue)
                    })

                    script.innerHTML = el.innerHTML;
                    el.replaceWith(script)
                }
            },

            key: (el) => {
                if (isntElement(el)) return

                return el.hasAttribute(`silica:key`)
                    ? el.getAttribute(`silica:key`)
                    : // If no "key", then first check for "silica:id", then "id"
                    el.hasAttribute(`silica:id`)
                        ? el.getAttribute(`silica:id`)
                        : el.id
            },

            lookahead: false,
        })
    }

    updateDomWithMorphdom(html) {
        const targetElement = document.querySelector('[silica\\:id="' + this.id + '"]');

        if (targetElement) {
            // Create a temporary div to hold the new HTML content
            let tempDiv = document.createElement("div");
            tempDiv.innerHTML = html;

            // Use morphdom to update the target element with the new content
            morphdom(targetElement, tempDiv.firstChild, {
                getNodeKey: function (node) {
                    if (typeof node.hasAttribute !== 'function') {
                        return
                    }

                    console.log(node.hasAttribute('silica:id') ? node.getAttribute('silica:id') : null)

                    return node.hasAttribute('silica:key')
                        ? node.getAttribute('silica:key')
                        : node.id
                },
                onNodeAdded: function (node) {
                    if (node.nodeName === 'SCRIPT') {
                        var script = document.createElement('script');
                        //copy over the attributes
                        [...node.attributes].forEach(attr => {
                            script.setAttribute(attr.nodeName, attr.nodeValue)
                        })

                        script.innerHTML = node.innerHTML;
                        node.replaceWith(script)
                    }
                },
                onBeforeElUpdated: (fromEl, toEl) => {
                    // If the element being updated is an input, ignore it
                    // if (fromEl.tagName === "INPUT" && toEl.tagName === "INPUT") {
                    //     return false;
                    // }

                    // Ignore glued items
                    if (
                        fromEl.hasAttribute("silica:glued") &&
                        toEl.hasAttribute("silica:glued")
                    ) {
                        return false;
                    }

                    // Ignore child components
                    if (fromEl.hasAttribute("silica:id") && toEl.hasAttribute("silica:id") && fromEl.getAttribute("silica:id") !== this.id) {
                        return false;
                    }

                    if (fromEl.nodeName === "SCRIPT" && toEl.nodeName === "SCRIPT") {
                        var script = document.createElement('script');
                        //copy over the attributes
                        [...toEl.attributes].forEach(attr => {
                            script.setAttribute(attr.nodeName, attr.nodeValue)
                        })

                        script.innerHTML = toEl.innerHTML;
                        fromEl.replaceWith(script)
                        return false;
                    }
                    return true; // Continue with the update for other elements
                }
            });

            this.removeListeners();
            this.setListeners();
        } else {
            console.warn(`Element with id="${this.id}" and name ${this.name} not found.`);
        }
    }

    processRedirections(jsCalls = []) {
        // Check if we have a redirect
        const redirectFn = jsCalls.find((call) => call.fn === "_silicaRedirect");
        if (redirectFn) {
            window.location.href = redirectFn.args[0];
            return true;
        }
        return false;
    }

    /**
     * Get all elements that match a query, but only if they are inside the subject component
     * @param query
     * @returns []
     */
    querySelectorAll(query) {
        let nodes = [];
        const nodeList = Array.from(this.el.querySelectorAll(query));

        for (let i = 0; i < nodeList.length; i++) {
            if (
                Silica.getNearestComponent(nodeList[i])?.name === this.name
            ) {
                nodes.push(nodeList[i]);
            }
        }
        return nodes;
    }

    /**
     * Show any silica:loading elements
     * @param action
     */
    showLoaders(action) {
        let target_from_action = null;

        if (action?.type === this.ACTION_TYPE_CALL_METHOD) {
            target_from_action = action?.method_name;
        } else if (action?.type === this.ACTION_TYPE_SET_PROPERTY) {
            target_from_action = action?.name;
        }

        this.querySelectorAll(
            "[silica\\:loading], [silica\\:loading\\.class]"
        ).forEach((el) => {
            if (el.hasAttribute("silica:target")) {
                const silica_target = el.getAttribute("silica:target");
                const target_without_parentheses = silica_target.split("(")[0];

                if (
                    [silica_target, target_without_parentheses].includes(
                        el.getAttribute("silica:target")
                    )
                ) {
                    el.style.display = "block";

                    const classAttr = el.getAttribute("silica:loading.class");
                    this.applyClassesToEl(el, classAttr);
                }
            } else {
                el.style.display = "block";

                const classAttr = el.getAttribute("silica:loading.class");
                this.applyClassesToEl(el, classAttr);
            }
        });
    }

    hideLoaders(action) {
        let target_from_action = null;

        if (action?.target === this.ACTION_TYPE_CALL_METHOD) {
            target_from_action = action?.method_name;
        } else if (action?.target === this.ACTION_TYPE_SET_PROPERTY) {
            target_from_action = action?.name;
        }

        this.querySelectorAll("[silica\\:loading]").forEach((el) => {
            if (el.hasAttribute("silica:target")) {
                const silica_target = el.getAttribute("silica:target");
                const target_without_parentheses = silica_target.split("(")[0];

                if (
                    [silica_target, target_without_parentheses].includes(
                        el.getAttribute("silica:target")
                    )
                ) {
                    el.style.display = "none";

                    const classAttr = el.getAttribute("silica:loading.class");
                    this.removeClassesFromEl(el, classAttr);
                }
            } else {
                el.style.display = "none";

                const classAttr = el.getAttribute("silica:loading.class");
                this.removeClassesFromEl(el, classAttr);
            }
        });
    }

    applyClassesToEl(el, classes = "") {
        if (classes) {
            // Split by space to support multiple classes
            const classes_arr = classes.split(" ");
            el.classList.add(...classes_arr);
        }
    }

    removeClassesFromEl(el, classes = "") {
        if (classes) {
            // Split by space to support multiple classes
            const classes_arr = classes.split(" ");
            el.classList.remove(...classes_arr);
        }
    }
}